/*************************************************************************************
Author:	Mayec Rancel
http://www.mayec.eu

Script Version:	0.2b4

Tested for Max Versions: 	max2009 x64. Won't work in versions previous to 9 because this script requires dotnet support.

Purpose:		Changes same parameter to same value in all selected objects (in which property exists)

Known Bugs:

To Do:

	Primary:
	- When selecting a parameter, read actual value and display it in field
	- add point3 fields
	- add transform channels

	
	Secondary:
	- be able to set controller's parameters (if controller type is shared)
	- create controller types arrays procedurally (now they are declared explictly)
	- button "Check", which returns if property is present, and in how many of all selected objects (and checks type of entered value against propertie's type)
	- hability to set several parameters and then apply
	- save last parameter and value entered and use them as default.
	- Add support for Complex (list) Controllers.
*************************************************************************************/

macroScript massParamChange
category:"MayecTools"
buttontext:"massParamChanger"
tooltip:"Mass Parameter Changer"
Icon:#("Maxscript",3)

(
	version = "0.2b4"
	-- make rollout global to allow use of callbacks
	global massParamChangeRollout
	try(destroyDialog massParamChangeRollout)catch() 
	
	-- declare arrays of possible Constraint types for the 3 transforms (position, rotation, scale)
	global posContTypes = #("linear_position","bezier_position","Path_Constraint","Position_Expression","Position_Constraint","tcb_position","Position_XYZ","Motion_Clip_SlavePos","Position_Layer","Sunlight_Daylight_Slave_ControllerPositionControlle","position_list","AudioPosition","position_script","Position_Wire","IK_Position_Controller","SlavePos","Dynamics_Position_Controller","Position_Mixer_Controller","Sunlight_Daylight_Slave_Controller","SpringPositionController","Position_Reactor","Noise_position","Attachment","Surface_position","position_ListDummyEntry","Position_Motion_Capture","Missing_Position_Control")
	global rotationContTypes = #("linear_rotation","bezier_rotation","Euler_XYZ","Local_Euler_XYZ","Orientation_Constraint","LookAt_Constraint","tcb_rotation","Motion_Clip_SlaveRotation","Expose_Euler_Control","Rotation_Reactor","rotation_script","Rotation_Wire","rotation_list","Dynamics_Rotation_Controller","AudioRotation","Rotation_Mixer_Controller","SlaveRotation","Rotation_Layer","Noise_rotation","rotation_ListDummyEntry","Rotation_Motion_Capture","Missing_Rotation_Control")
	global scaleContTypes = #("linear_scale","bezier_scale","Scale_Expression","tcb_scale","ScaleXYZ","Scale_Reactor","Scale_Layer","AudioScale","scale_list","Scale_Mixer_Controller","SlaveScale","Motion_Clip_SlaveScale","scale_script","Scale_Wire","Noise_scale","scale_ListDummyEntry","Scale_Motion_Capture","Missing_Scale_Control")

	-- finds common elements in arrays a and b
	fn arrayIntersect theArrays =
	(
		intersection = #()
		a = theArrays[1]
		b = theArrays[2]
		for elemt in a do
				if ( (findItem b elemt) != 0 ) then append intersection elemt
		if theArrays.count > 2 then (
			for n = 3 to theArrays.count do (
				intersection = arrayIntersect #(intersection,theArrays[n])
			)
		)
		return intersection
	)

	-- function returns the number of common modifiers of theNodes (by comparing modifier names and order):
	fn commonMods theNodes =
	(
		if theNodes[1].modifiers.count > 0 then
		(
			intCommonMods = theNodes[1].modifiers.count
			for n = 2 to theNodes.count do
			(
				if intCommonMods > theNodes[n].modifiers.count then intCommonMods = theNodes[n].modifiers.count
				newCount = intCommonMods
				for m = 1 to intCommonMods do
				(
					origIndex = theNodes[1].modifiers.count-intCommonMods+1 -- we have to invert the order of modifier checking (last in stack has lowest index)
					modIndex = theNodes[n].modifiers.count-intCommonMods+1 -- we have to invert the order of modifier checking (last in stack has lowest index)
					origName = theNodes[1].modifiers[origIndex].name
					modName = theNodes[n].modifiers[modIndex].name
					if modName == origName then newCount = m
					else
						if m == newCount then newCount = 0
				)
				intCommonMods = newCount
			)
			return intCommonMods
		)
		else return 0
	)

	-- function returns an array with the properties present in all theNodes, at base level (uses function arrayIntersect)
	fn commonBaseProperties theNodes =
	(
		propArrays = #()
		for n in theNodes do
			append propArrays (getPropNames n)
		result = arrayIntersect propArrays
		return result
	)

	-- function checks if a property exists in a node (considers also modifier properties expressed as "modifier.property")
	fn checkProperty theNode theProp =
	(
		propLevels = filterString theProp "."
		-- if no more than 1 property levels, do a straight isProperty check
		if propLevels.count < 2 then
		(
			if (isProperty theNode theProp) then return true
			else return false
		)
		-- else, check each sub-node with successive isProperty applications
		else (
			for l = 1 to propLevels.count do
			(
				if ( (isProperty theNode propLevels[l]) == false ) do return false
				if ( l < propLevels.count ) do theNode = theNode.modifiers[propLevels[l]]
			)
			return true
		)
	)

	rollout massParamChangeRollout ("Mass Parameter Changer "+version) width:472 height:429
	(
		-- Functions for comoon parameters TreeView
		fn initTreeView tv = (tv.Indent = 14)
			
		fn fillInTreeView tv =
		(	
			tv.Nodes.clear()
			allObjects = for o in selection collect o
			if allObjects.count > 1 then
			(
				-- analyze how many common modifiers the selected objects have
				commonModifiers = commonMods allObjects
				comProperties = commonBaseProperties allObjects
				baseNode = tv.Nodes.add "base" -- add parent node
				-- populate tree view with common properties
				for p = 1 to comProperties.count do
				(
					newNode = baseNode.Nodes.add (comProperties[p] as string)
					--newNode.tag = newNode.text
					newNode.name = newNode.text
				)
				-- populate tree view with common modifiers (if any)
				if commonModifiers > 0 then
				(
					modsNode = tv.Nodes.add "modifiers"
					-- run through common modifiers
					for m = 1 to commonModifiers do
					(
						index = allObjects[1].modifiers.count - m + 1
						theModNode = modsNode.Nodes.add (allObjects[1].modifiers[index].name as string)
						-- run through modifier's properties
						theModProps = getPropNames allObjects[1].modifiers[index]
						for mp = 1 to theModProps.count do
						(
							newNode = theModNode.Nodes.add (theModProps[mp] as string)
							--newNode.tag =  (theModNode.text + "." + newNode.text)
							newNode.name =  (theModNode.text + "." + newNode.text)
						)
					)
				)
				-- populate tree view with CONTROLLERS
				contNode = tv.Nodes.add "controllers"
				newNode = contNode.Nodes.add ("position")
				newNode.name = "pos.controller"
				newNode = contNode.Nodes.add ("rotation")
				newNode.name = "rotation.controller"
				newNode = contNode.Nodes.add ("scale")
				newNode.name = "scale.controller"
				-- expand tree view
				baseNode.expand()
				if commonModifiers > 0 then modsNode.expand()
				contNode.expand()
			)
		)
		
		edittext edtParameter "Type parameter name:" pos:[15,16] width:344 height:19
		--combobox cbxParameter "or select Parameter:" pos:[15,43] width:180 height:21 
		-- Common Parameters TreeView
		label labelTreeView "...or select parameter:" pos:[15,48]
		
		label labelValInput "New Value:" pos:[213,61] width:246 height:19
		edittext edtStrValue "String:" pos:[213,91] width:246 height:19
		on edtStrValue changed foo do (global theNewValue = edtStrValue.text)
		label labelP3note "To define a point3 value, type 3 numbers separated by commas. e.g. '2,3.6,5' " pos:[213,121] width:246 height:40
		checkbox chkBoolValue "True / False" pos:[223,161] enabled:false
		on chkBoolValue changed foo do (global theNewValue = chkBoolValue.checked)
		spinner spnIntValue "Integer :" pos:[245,191] width:106 type:#integer enabled:false range:[-2147483647,214748369,0]
		on spnIntValue changed foo do (global theNewValue = spnIntValue.value)
		spinner spnFloatValue "Float :" pos:[245,221] width:106 type:#float enabled:false range:[-3.40E38,3.40E38,0]
		on spnFloatValue changed foo do (global theNewValue = spnFloatValue.value)
		dropdownlist ddlContValue "Controller type:" pos:[225,251] width:236 enabled:false
		on ddlContValue selected foo do (global theNewValue = ddlContValue.selected)
		
		dotNetControl tv "TreeView" pos:[15,63] width:190 height:350 align:#center
		-- initialize treeView
		on massParamChangeRollout open do
		(
			initTreeView tv 
			fillInTreeView tv
		)	
		-- what to do when a node is selected in TreeView
		on tv click arg do
		(
			--First get the TreeView node below the mouse cursor
			--The arg argument has properties .x and .y with the current pos.
			--Use showProperties arg to see what is available...
			--We use the TreeView method GetNodeAt to see what was clicked:
			hitNode = tv.GetNodeAt (dotNetObject "System.Drawing.Point" arg.x arg.y)
			if hitNode != undefined do --if a TreeView node was clicked,
				try(edtParameter.text = hitNode.name)catch(print "treeview selection not a parameter")
			if (filterString hitNode.name ".")[2] == "controller" then
			(
				ddlContValue.enabled = true
				print "selected a Controller type"
				case hitNode.name of
				(
					"pos.controller" :
						ddlContValue.items = posContTypes
					"rotation.controller":
						ddlContValue.items = rotationContTypes
					"scale.controller":
						ddlContValue.items = scaleContTypes
				)
			)
			-- check if its a valid node or a headqer
			else if checkProperty selection[1] hitNode.name then
			(
				if (filterString hitNode.name ".").count < 2 then
				(
					theCompNode = selection[1]
					global theClass = classOf (getProperty theCompNode hitNode.name)
				)
				else
				(
					brokenDownName = filterString hitNode.name "."
					theCompNode = selection[1].modifiers[brokenDownName[1]]
					theClass = classOf (getProperty theCompNode brokenDownName[2] )
				)
				case theClass of
				(
					Integer :
					(
						print "selected property is of Integer type"
						chkBoolValue.enabled = false
						spnIntValue.enabled = true
						spnFloatValue.enabled = false
						ddlContValue.enabled = false
						theNewValue = spnIntValue.value
					)
					Float :
					(
						print "selected property is of Float type"
						chkBoolValue.enabled = false
						spnIntValue.enabled = false
						spnFloatValue.enabled = true
						ddlContValue.enabled = false
						theNewValue = spnFloatValue.value
					)
					BooleanClass :
					(
						print "selected property is of Boolean type"
						chkBoolValue.enabled = true
						spnIntValue.enabled = false
						spnFloatValue.enabled = false
						ddlContValue.enabled = false
						theNewValue = chkBoolValue.checked
					)
					String :
					(
						print "selected property is of String type"
						chkBoolValue.enabled = false
						spnIntValue.enabled = false
						spnFloatValue.enabled = false
						ddlContValue.enabled = false
						theNewValue = edtStrValue.text
					)
					Point3 :
					(
						print "selected property is of Point3 type"
						chkBoolValue.enabled = false
						spnIntValue.enabled = false
						spnFloatValue.enabled = false
						ddlContValue.enabled = false
						theNewValue = edtStrValue.text
					)
					
					default:
					(
						print "selected property is of unrecognized type"
						chkBoolValue.enabled = false
						spnIntValue.enabled = false
						spnFloatValue.enabled = false
						ddlContValue.enabled = false
						theNewValue = edtStrValue.text
					)
				)
			)
		)
		
	-- Go
		button btnGo "Apply!" pos:[380,329] width:80 height:29
		
	-- Close
		button btnClose "Close" pos:[250,329] width:80 height:29
		
	-- About
		button btnAbout "About" pos:[324,374] width:70 height:22
		
		-- when dialog is closed, remove callback that updates the "selected objects" list
		on massParamChangeRollout close do callbacks.removeScripts #selectionSetChanged id:#massParamChange
		
		on btnGo pressed do
		(
			-- check if there are inputs for both a parameter and its value
			if edtParameter.text != "" and theNewValue != "" then
			(
				
				-- if property is a controller treat differently:
				if (filterString edtParameter.text ".")[2] == "controller"  then
				(
					theChannel = (filterString edtParameter.text ".")[1]
					for obj in selection do
					(
						case theChannel of
						(
							"pos":
								obj.pos.controller = (execute (theNewValue + "()") )
							"rotation": 
								obj.rotation.controller = (execute (theNewValue + "()") )
							"scale":
								obj.scale.controller = (execute (theNewValue + "()") )
						)
					)
				)
				
				-- if it's not a controller...
				else (
					-- Check if property exists in at least one of the selected objects
					foundQuant = 0
					for obj in selection do
					(
						if (checkProperty obj edtParameter.text) then (foundQuant = foundQuant + 1)
					)
					if foundQuant > 0 then (
						for obj in selection do
						(
							
							-- parse edtParameter.text  to see if its a simple property or a modifier property
							nodeLevels = filterString edtParameter.text "."
							if nodeLevels.count < 2 then (
								theCompNode = obj
								theProperty = edtParameter.text
							)
							else (
								theCompNode = obj.modifiers[nodeLevels[1]]
								theProperty = nodeLevels[2]
							)
							
							case theClass of
							(
								Integer :
									setProperty theCompNode theProperty  (theNewValue as integer)
								Float :
									setProperty theCompNode theProperty  (theNewValue as float)
								BooleanClass :
									setProperty theCompNode theProperty  (theNewValue as booleanClass)
								Point3 :
								(
									valToP3 = filterString theNewValue ","
									newP3Val = point3 (valToP3[1] as float) (valToP3[2] as float) (valToP3[3] as float)
									setProperty theCompNode theProperty newP3Val
								)
								default:
								(
									-- if "true" or "false" value was entered, treat as boolean
									if ( (theNewValue == "true") or (theNewValue == "false") ) then
										setProperty theCompNode theProperty  (theNewValue as booleanClass)
									-- else treat as a number
									else
									setProperty theCompNode theProperty  (theNewValue as number)
								)
							)
							
						)
					)
					else (messagebox ("ERROR: none of the selected objects has a ." + edtParameter.text + " property") )
				)

			) -- close check for existing inputs
			else	messagebox ("You need to enter a parameter and its new value")
		)
		on btnAbout pressed do
		(
			messagebox ("Changes same parameter to same value in all the selected objects (in which the specific property exists)\n\nVersion: "+version+"\n\nWritten by \nMayec Rancel \nhttp://www.mayec.eu")
		)
		on btnClose pressed do
			destroyDialog massParamChangeRollout
	)

	createDialog massParamChangeRollout
	-- Add callback to update treeView on selection change
	callbacks.addScript #selectionSetChanged "massParamChangeRollout.fillInTreeView massParamChangeRollout.tv" id:#massParamChange

)